/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package vehicle;

import com.jme3.math.Quaternion;
import com.jme3.math.Vector3f;
import com.jme3.renderer.RenderManager;
import com.jme3.renderer.ViewPort;
import com.jme3.scene.Node;
import com.jme3.scene.control.AbstractControl;

/**
 *
 * @author Eirik
 */
public class InverseKinematics extends AbstractControl {

    private Node[] bones;   //  First node is a static baseNode, last node is a point at the very end of your chain
    private Vector3f targetPosition = Vector3f.ZERO;
    private boolean active;
    private boolean rotationLocked;

    public InverseKinematics(Node[] bones, boolean rotationLocked) {
        this.bones = bones;
        this.rotationLocked = rotationLocked;
    }

    @Override
    public void controlUpdate(float tpf) {

        if (bones.length == 0 || !active) {
            return;
        }

        int iterations = 0;
        for (int i = bones.length - 2; i > 0; i--) {
            Vector3f e = getEffectorPosition();                 // effector
            Vector3f j = bones[i].getWorldTranslation().clone();// current join position
            Vector3f t = targetPosition;                        //target
            if (rotationLocked) {
                Vector3f pos = bones[0].getWorldTranslation();  //  Get base position
                Quaternion rot = bones[0].getWorldRotation().inverse(); //  Get rotation
                t = rot.mult(t.subtract(pos));    //  transform to localspace
                t = new Vector3f(0, t.y, t.z); //   Remove axis
                t = rot.inverse().mult(t);  //  transform to worldspace
                t = pos.add(t);
            }

            Vector3f currentDir = e.subtract(j).normalizeLocal();
            Vector3f target = t.subtract(j).normalizeLocal();

            float angle = currentDir.angleBetween(target) * tpf * 30f;
            Vector3f cross = currentDir.cross(target);

            Quaternion q = new Quaternion().fromAngleAxis(angle, cross);
            Quaternion targetWorld = q.mult(bones[i - 1].getWorldRotation());
            Quaternion diff = bones[i - 1].getWorldRotation().inverse().mult(targetWorld);
            bones[i].setLocalRotation(diff.mult(bones[i].getLocalRotation()));
        }
    }

    private Vector3f getEffectorPosition() {   //  end pivot
        return bones[bones.length - 1].getWorldTranslation();
    }

    public void setTarget(Vector3f tar) {
        targetPosition = tar;
    }

    public void setActive(boolean active) {
        this.active = active;
    }

    @Override
    protected void controlRender(RenderManager rm, ViewPort vp) {

    }
}
