/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package vehicle;

import a_program.Program;
import static a_program.Program.assetManager;
import static a_program.Program.bulletAppState;
import c_scenario.ScenarioTriggerTrackPos;
import c_trackNode.TrackRunner;
import com.jme3.asset.AssetManager;
import com.jme3.bullet.BulletAppState;
import com.jme3.bullet.collision.shapes.BoxCollisionShape;
import com.jme3.bullet.collision.shapes.CompoundCollisionShape;
import com.jme3.bullet.collision.shapes.MeshCollisionShape;
import com.jme3.bullet.control.RigidBodyControl;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.FastMath;
import com.jme3.math.Matrix3f;
import com.jme3.math.Quaternion;
import com.jme3.math.Vector3f;
import com.jme3.scene.CameraNode;
import com.jme3.scene.Geometry;
import com.jme3.scene.Node;
import com.jme3.scene.Spatial;
import com.jme3.scene.shape.Box;
import com.jme3.util.TangentBinormalGenerator;
import e_tco.Action3dElement;
import e_tco.GeometryAction;
import f_general.MyAudioNode;
import java.util.ArrayList;
import e_tco.TrackConnectedObject;
import f_general.AnglesVectors;
import vehicle.VehicleData.VehicleType;

/**
 *
 * @author hallgeo
 */
public class Vehicle {

    private String id;
    private TrackRunner t1;
    private TrackRunner t2;

    protected float speedKmh = 0;
    private float speedMs = speedKmh / 3.6f;
    protected float odometer = 100;
    public static float secsSinceLastFrame = 0.0167f; // blir oppdatert av simpleUpdate()
    private float metersSinceLastFrame = 0;
    private float yPos = 0f;

    private Vector3f vehicleDirectionVector = new Vector3f();
    private Vector3f vehicleLocationVector = new Vector3f();
    private Vector3f lookAtVector = new Vector3f();
    private Vector3f lookAtVectorReverse = new Vector3f();
    private float angleAroundY;
    private float angleAroundPlain;
    private float angleAroundZ = 0.0f;
    public float stigningsverdi;
    private float angleAroundYPrev;
    private float angleAroundPlainPrev;
    private float angleAroundZPrev;
    private float rotDiffY;
    private float rotDiffYPrev;
    private float rotDiffPlain;
    private float rotDiffPlainPrev;

    protected Node pivot;
    protected Node bogie1Pivot = new Node();
    protected Node bogie2Pivot = new Node();
    protected Node guardrail1Pivot = new Node();
    protected Node guardrail2Pivot = new Node();
    protected ArrayList<Spatial> wheels = new ArrayList<Spatial>();
    protected ArrayList<Spatial> brakes = new ArrayList<Spatial>();
    protected float brakeOffset = .05f;
    private float lookAtDistance = 1000f;

    // boggivinkler
    private Vector3f boggi1Vec = new Vector3f();
    private Vector3f boggi2Vec = new Vector3f();
    private float boggi1OmY = 0;
    private float boggi2OmY = 0;
    private float boggi_1_2_Diff = 0;

    // connected vehicles
    protected Vehicle vehicle1;
    protected Vehicle vehicle2;
    protected boolean vehicle1Connected = false;
    protected boolean vehicle2Connected = false;
    protected float vehicleOdomOffset = 0.0f;

    protected int direction = 1;
    protected boolean noUpdate = false;

    // audio
    protected ArrayList<MyAudioNode> vehicleAudioList = new ArrayList<MyAudioNode>();

    // camera
    protected CameraNode camNode;

    // spoling
    protected boolean spolingOn = false;
    protected float spolingSpeedKmh = 0;
    protected float spolingSpeedMs = 0;
    //protected Spatial observer;
    protected boolean harStigtrinn = true;

    Vehicle pushing;
    Vehicle pushedBy;

    Coupling pushingCoupling;
    Coupling pushedByCoupling;

    // Couplings
    protected Coupling coupling1;
    protected Coupling coupling2;
    protected AirCoupling air1;
    protected AirCoupling air2;
    
    protected VehicleBrakeSystem vbs;

    protected Guardrail guardrail1;
    protected Guardrail guardrail2;

    VehicleData vd;
    String vehicleName = "locomotive";
    protected RigidBodyControl rb;

    protected boolean locomotive = false;

    float brake = 0;

    public Vehicle(String vehicleName) {
        this.vehicleName = vehicleName;
        init();
        init3DVehicle(null);
    }

    public Vehicle(String vehicleName, String ac3dFile) {
        this.vehicleName = vehicleName;
        init();
        init3DVehicle(ac3dFile);
    }

    public String getId() {
        return id;
    }

    public void setId(String id) {
        this.id = id;
    }

    public float getOdometer() {
        return odometer;
    }

    public void setOdometer(float odometer) {
        this.odometer = odometer;
    }

    public void setSpeedKmh(float speedKmh) {
        this.speedKmh = speedKmh;
        this.speedMs = speedKmh / 3.6f;
    }

    public float getSpeedKmh() {
        return speedKmh;
    }

    public float getSpeedMs() {
        return speedMs;
    }

    public void setBrake(float b) {
        brake = b;
    }

    public float getBrake() {
        return brake;
    }

    public void setyPos(float yPos) {
        this.yPos = yPos;
    }

    public CameraNode getCamNode() {
        return camNode;
    }

    public void setCamNode(CameraNode camNode) {
        this.camNode = camNode;
    }

    public ArrayList<MyAudioNode> getVehicleAudioList() {
        return vehicleAudioList;
    }

    public void changeToCamCab2() {
        camNode.setLocalTranslation(new Vector3f(-0.7f, 1.0f, 10));

        // snu kameraet
        Quaternion q1 = new Quaternion();
        q1.fromAngleAxis(0, new Vector3f(0, 1, 0));
        camNode.setLocalRotation(q1);
    }

    public void changeToCamCab1() {
        camNode.setLocalTranslation(new Vector3f(0.7f, 1.0f, -10));

        // snu kameraet
        Quaternion q1 = new Quaternion();
        q1.fromAngleAxis(FastMath.DEG_TO_RAD * 180, new Vector3f(0, 1, 0));
        camNode.setLocalRotation(q1);
    }

    public void tick1(/*float odom*/) {
        //setOdometer(odom - halfLength);
        //tick();

        if (coupling1.getConnectedTo() != null) {
            coupling1.getConnectedTo().getVehicle().setSpeedKmh(speedKmh * 1);
            coupling1.getConnectedTo().getVehicle().setDirection(direction);
            coupling1.getConnectedTo().getVehicle().tick1();
        } else if (air1.getConnectedTo() != null && this.speedKmh > 0
                && ((direction > 0 && speedKmh > air1.getConnectedTo().getVehicle().getSpeedKmh())
                || (direction <= 0 && speedKmh < air1.getConnectedTo().getVehicle().getSpeedKmh()))) {
            air1.disconnect();// Disconnects air if it is connected to a cart going away from it
        }                            // Should do something else in the future, maybe disconnecting violently  
    }

    public void tick2(/*float odom*/) {
        //setOdometer(odom + halfLength);
        //tick();

        if (coupling2.getConnectedTo() != null) {
            coupling2.getConnectedTo().getVehicle().setSpeedKmh(speedKmh * 1);
            coupling2.getConnectedTo().getVehicle().setDirection(direction);
            coupling2.getConnectedTo().getVehicle().tick2(/*odometer + halfLength + 0.75f*/);
        } else if (air2.getConnectedTo() != null && this.speedKmh > 0
                && ((direction <= 0 && speedKmh > air2.getConnectedTo().getVehicle().getSpeedKmh())
                || (direction < 0 && speedKmh < air2.getConnectedTo().getVehicle().getSpeedKmh()))) {
            air2.disconnect();    // Disconnects air if it is connected to a cart going away from it   
        }
    }

    public void tickMain() {
        if (noUpdate) {
            return;
        }

        //tick();
        if (coupling1.getConnectedTo() != null) {
            if (direction > 0 && speedKmh > coupling1.getConnectedTo().getVehicle().getSpeedKmh()) {   // If dragging a cart
                coupling1.getConnectedTo().getVehicle().setSpeedKmh(speedKmh);
                coupling1.getConnectedTo().getVehicle().setDirection(direction);
                coupling1.getConnectedTo().getVehicle().tick1(/*odometer + 0.5f*/);
            } else if (direction <= 0 && speedKmh < coupling1.getConnectedTo().getVehicle().getSpeedKmh()) {   // If braking a cart
                coupling1.getConnectedTo().getVehicle().setSpeedKmh(speedKmh);
                coupling1.getConnectedTo().getVehicle().setDirection(direction);
                coupling1.getConnectedTo().getVehicle().tick1(/*odometer + 0.5f*/);
            }
        } else if (air1.getConnectedTo() != null && this.speedKmh > 0
                && ((direction > 0 && speedKmh > air1.getConnectedTo().getVehicle().getSpeedKmh())
                || (direction <= 0 && speedKmh < air1.getConnectedTo().getVehicle().getSpeedKmh()))) {
            air1.disconnect();    // Disconnects air if it is connected to a cart going away from it
        }                            // Should do something else in the future, maybe disconnecting violently                   
        if (coupling2.getConnectedTo() != null) {
            if (direction <= 0 && speedKmh > coupling2.getConnectedTo().getVehicle().getSpeedKmh()) {   // If dragging a cart
                coupling2.getConnectedTo().getVehicle().setSpeedKmh(speedKmh * 1);
                coupling2.getConnectedTo().getVehicle().setDirection(direction);
                coupling2.getConnectedTo().getVehicle().tick2(/*odometer - 0.5f*/);
                //this.setSpeedKmh(speedKmh*0.9f);
            } else if (direction > 0 && speedKmh < coupling2.getConnectedTo().getVehicle().getSpeedKmh()) {   // If braking a cart
                coupling2.getConnectedTo().getVehicle().setSpeedKmh(speedKmh);
                coupling2.getConnectedTo().getVehicle().setDirection(direction);
                coupling2.getConnectedTo().getVehicle().tick2(/*odometer + 0.5f*/);
            }
        } else if (air2.getConnectedTo() != null && this.speedKmh > 0
                && ((direction <= 0 && speedKmh > air2.getConnectedTo().getVehicle().getSpeedKmh())
                || (direction < 0 && speedKmh < air2.getConnectedTo().getVehicle().getSpeedKmh()))) {
            air2.disconnect();    // Disconnects air if it is connected to a cart going away from it
        }
    }

    private Vector3f vehicleLocationVectorPrev = new Vector3f();

    public void tick() {
        coupling1.tick(secsSinceLastFrame); // Update with tpf
        coupling2.tick(secsSinceLastFrame); // Update with tpf
        air1.tick(secsSinceLastFrame); // Update with tpf
        air2.tick(secsSinceLastFrame); // Update with tpf
        vbs.tick(secsSinceLastFrame); // Update with tpf
        
        // Spin wheels
        for (Spatial w : wheels) {
            w.rotate(speedMs * direction * .1f, 0, 0);
        }

        if (!spolingOn) // normal
        {
            metersSinceLastFrame = speedMs * secsSinceLastFrame * direction;
        } else // spoling mode
        {
            metersSinceLastFrame = spolingSpeedMs * secsSinceLastFrame;
        }

        odometer += metersSinceLastFrame;

        t1.tick(odometer - vd.trainColliderLength);
        t2.tick(odometer + vd.trainColliderLength);

        //
        vehicleDirectionVector.x = t2.getPoint().x - t1.getPoint().x;
        vehicleDirectionVector.y = t2.getPoint().y - t1.getPoint().y;
        vehicleDirectionVector.z = t2.getPoint().z - t1.getPoint().z;

        //
        vehicleLocationVector.x = (t2.getPoint().x + t1.getPoint().x) / 2;
        vehicleLocationVector.y = ((t2.getPoint().y + t1.getPoint().y) / 2) + yPos;
        vehicleLocationVector.z = (t2.getPoint().z + t1.getPoint().z) / 2;

        findAngles();

        if (!Program.vrOn) {
            pivot.setLocalTranslation(vehicleLocationVector);
        } else {
            // BRUKER NÅ PREV I VR OR Å FORFLYTTE TOGET PÅ FORRIGE POSISJON
            pivot.setLocalTranslation(vehicleLocationVectorPrev);
        }

        rotDiffY = angleAroundY - angleAroundYPrev;
        rotDiffPlain = angleAroundPlain - angleAroundPlainPrev;

        pivot.rotate(0, rotDiffY, 0);
        pivot.rotate(-rotDiffPlain, 0, 0);

        angleAroundYPrev = angleAroundY;
        angleAroundPlainPrev = angleAroundPlain;

        // store prevs
        vehicleLocationVectorPrev.x = vehicleLocationVector.x;
        vehicleLocationVectorPrev.y = vehicleLocationVector.y;
        vehicleLocationVectorPrev.z = vehicleLocationVector.z;

        // brukes ikke pr nå
        rotDiffYPrev = rotDiffY;
        rotDiffPlainPrev = rotDiffPlain;

        if (pushingCoupling != null
                && pushingCoupling.getBasePivot().getWorldTranslation().distance(pushingCoupling.getVehicle().getPushedBy().getBasePivot().getWorldTranslation()) > 0.85f) {
            pushingCoupling.getVehicle().setPushedBy(null);
            pushingCoupling = null;
        }

        if (!locomotive) {
            if (pushedByCoupling == null) {
                setSpeedKmh(speedKmh * 0.9985f);
            }

            setSpeedKmh(speedKmh - (speedKmh * brake / 100));

            if (speedKmh < 1.5f && pushedByCoupling == null) {
                setSpeedKmh(0);
            }
        }
    }

    public void collision() {
        if (speedKmh < 0.15f) {
            return;
        }

        Coupling coupling;
        float spacing = vd.trainColliderLength;

        if (direction > 0) {
            coupling = getCoupling(1);
        } else {
            coupling = getCoupling(0);
        }

        for (Vehicle v : Program.stVehicleList) {
            if (v != coupling.getVehicle()) {
                for (int i = 0; i < 2; i++) {
                    if (v.getCoupling(i).getBasePivot().getWorldTranslation().distance(coupling.getBasePivot().getWorldTranslation()) < 0.83f) {
                        if (v.getSpeedKmh() < speedKmh) {
                            float totalSpeed = speedKmh + v.getSpeedKmh();
                            float dist = this.odometer - v.getOdometer();
                            System.out.println(dist);
                            spacing += v.vd.trainColliderLength;
                            if (direction <= 0)
                                spacing *= -1;
                            v.setOdometer(odometer + spacing);

                            v.setSpeedKmh(totalSpeed / 2);
                            this.setSpeedKmh(totalSpeed / 2);

                            v.setDirection(direction);

                            this.pushingCoupling = v.getCoupling(i);
                            v.setPushedBy(coupling);
                        }
                        return;
                    }
                }
            }
        }
    }

    public void setPushedBy(Coupling c) {
        this.pushedByCoupling = c;
    }

    public Coupling getPushedBy() {
        return this.pushedByCoupling;
    }

    public float getMetersSinceLastFrame() {
        return metersSinceLastFrame;
    }

    public void setMetersSinceLastFrame(float metersSinceLastFrame) {
        this.metersSinceLastFrame = metersSinceLastFrame;
    }

    public Vector3f getVehicleLocationVector() {
        return vehicleLocationVector;
    }

    public void updateTickWhilePause() {
        // to be overridden
    }

    public void functionsUpdateCh1() {

    }

    public float findCurrentPosInSection() {
        float a = t2.getPosSection();
        float b = t1.getPosSection();

        return (a + b) / 2;
    }

    public void passTCO(TrackConnectedObject tco) {
        System.out.println("Vehicle passing TCO: " + tco.getClass());

        if (tco instanceof ScenarioTriggerTrackPos) {
            ((ScenarioTriggerTrackPos) tco).passedByVehicle(this);
        }
        //HelloAssets.pause();
    }

    public void pauseButton() {
        // not implemented
    }

    public TrackRunner getT1() {
        return t1;
    }

    public TrackRunner getT2() {
        return t2;
    }

    public TrackRunner getTrackRunner1() {
        return t1;
    }

    public TrackRunner getTrackRunner2() {
        return t2;
    }

    public Node getPivot() {
        return pivot;
    }

    public Vehicle getVehicle1() {
        return vehicle1;
    }

    public void setVehicle1(Vehicle vehicle1) {
        this.vehicle1 = vehicle1;
    }

    public Vehicle getVehicle2() {
        return vehicle2;
    }

    public void setVehicle2(Vehicle vehicle2) {
        this.vehicle2 = vehicle2;
    }

    public int getDirection() {
        return direction;
    }

    public void setDirection(int direction) {
        this.direction = direction;
    }

    public float getHalfLength() {
        return vd.trainColliderLength;
    }

    public void setVehicleOdomOffset(float vehicleOdomOffset) {
        this.vehicleOdomOffset = vehicleOdomOffset;
    }

    public void setNoUpdate(boolean noUpdate) {
        this.noUpdate = noUpdate;
    }

    private float angY = 0;
    private float angY2 = 0;
    private float angYD = 0;

    private void findAngles() {
        if (vehicleDirectionVector.z < 0) {
            angleAroundY = (FastMath.atan(vehicleDirectionVector.x / vehicleDirectionVector.z)) + (180 * FastMath.DEG_TO_RAD);
        } else {
            angleAroundY = (FastMath.atan(vehicleDirectionVector.x / vehicleDirectionVector.z));
        }

        stigningsverdi = vehicleDirectionVector.y / (2*vd.trainColliderLength);
        angleAroundPlain = (FastMath.asin(stigningsverdi));
    }

    public void setActiveTrackInfoReceiver(boolean a1, boolean a2) {
        t1.setTrackInfoReceiver(a1);
        t2.setTrackInfoReceiver(a2);
    }

    public void crashWithTrackRunnerNotice() {
        System.out.println("crash with trackrunner");
        Program.ibuttStopp();
    }

    protected void init() {
        
        // runner along track
        t1 = new TrackRunner(null, new Vector3f());
        t2 = new TrackRunner(null, new Vector3f());
        t1.setVehicle(this);
        t2.setVehicle(this);
        t1.setId(1);
        t2.setId(2);
        t1.setTrackInfoReceiver(false);
        t2.setTrackInfoReceiver(true);

        // pivot for 3d object
        pivot = new Node("pivot");
        
        if (vehicleName == "locomotive")
            vd = new VehicleData();
        else
            vd = new VehicleData(
                    "wagon", 
                    VehicleType.Wagon, 
                    1.4f, 
                    7.4f, 
                    7.9f, 
                    1.6f, 
                    8.4f, 
                    new Vector3f(0, 1.7f, 6.85f), 
                    new Vector3f(0, 1.26f, 7.91f), 
                    new Vector3f(-0.395f, 1.05f, 7.90f), 
                    new Vector3f(0, 0, 4.0f), 
                    0.0f, 
                    new Vector3f(0, .68f, 1.025f), 
                    new Vector3f(0,0,0));
    
    }

    protected void init3DVehicle(String ac3dFile) // ac3dfile not used here, but potentially in derived classes
    {
         String vehicleType = ""; 
        if (vd.vehicleType == VehicleType.Locomotive)
            vehicleType = "locomotive";
        if (vd.vehicleType == VehicleType.Wagon)
            vehicleType = "wagon";
        if (vd.vehicleType == VehicleType.LinkedWagon)
            vehicleType = "linkedWagon";
        
        Material matDebug = new Material(assetManager, "Common/MatDefs/Misc/ShowNormals.j3md");
        
        // connections to other vehicles
        addCoupling();
        addAirCoupling();
        
        float x;
        float y;
        float z;

        //  Base
        Spatial vehicleBase = loadModel("veh_" + vd.name);
        vehicleBase.setMaterial(matDebug);
        pivot.attachChild(vehicleBase);

        //  Collision shape
        CompoundCollisionShape shape = new CompoundCollisionShape();
        //  Base
        shape.addChildShape(new BoxCollisionShape(new Vector3f(vd.colliderWidth, 5f, vd.colliderLength2)), Vector3f.ZERO);
        /*//  Step
        shape.addChildShape(new BoxCollisionShape(new Vector3f(vd.colliderWidth / 2f, vd.colliderHeight, vd.colliderLength2)), Vector3f.ZERO);
        //  Steps
        shape.addChildShape(new BoxCollisionShape(new Vector3f(vd.colliderHeight / 1.9f, vd.colliderHeight / 2f, vd.colliderLength2)), new Vector3f(-.325f, .5f, 0), new Quaternion(0, 0, 1, FastMath.PI / 2f).toRotationMatrix());
        shape.addChildShape(new BoxCollisionShape(new Vector3f(vd.colliderHeight / 1.9f, vd.colliderHeight / 2f, vd.colliderLength2)), new Vector3f(.325f, .5f, 0), new Quaternion(0, 0, 1, -FastMath.PI / 2f).toRotationMatrix());
        //  Rails
        shape.addChildShape(new BoxCollisionShape(new Vector3f(vd.colliderWidth, vd.colliderHeight * 2f, .01f)), new Vector3f(0, 0, vd.colliderLength2));
        shape.addChildShape(new BoxCollisionShape(new Vector3f(vd.colliderWidth, vd.colliderHeight * 2f, .01f)), new Vector3f(0, 0, -vd.colliderLength2));*/
        //  Add Rigidbody to train, and add to physics space
        rb = new RigidBodyControl(shape, 100.0f);
        rb.setKinematic(true);
        pivot.addControl(rb);
        bulletAppState.getPhysicsSpace().add(rb);
        
        
        
        //  Guardrails
        if (vd.vehicleType == VehicleType.Locomotive)
        {
            guardrail1 = new Guardrail(this, guardrail1Pivot);
            guardrail2 = new Guardrail(this, guardrail2Pivot);
            pivot.attachChild(guardrail1Pivot);
            pivot.attachChild(guardrail2Pivot);
        
            x = vd.guardrailPos.x;
            y = vd.guardrailPos.y;
            z = vd.guardrailPos.z;
            guardrail1Pivot.rotate(0, FastMath.DEG_TO_RAD * 180, 0);
            guardrail1Pivot.setLocalTranslation(x, y, -z);
            guardrail2Pivot.setLocalTranslation(x, y, z);
        }

        //  Bogies
        x = vd.bogiePos.x;
        y = vd.bogiePos.y;
        z = vd.bogiePos.z;
        //  b1
        Spatial bogie1 = loadModel(vehicleType + "Bogie");
        //Spatial bogie1 = loadModel(assetManager, "wagonBogie");
        bogie1.setMaterial(matDebug);
        pivot.attachChild(bogie1Pivot);
        bogie1Pivot.attachChild(bogie1);
        bogie1Pivot.setLocalTranslation(x, y, z + vd.bogieOffset);
        //  b2
        Spatial bogie2 = bogie1.clone();
        pivot.attachChild(bogie2Pivot);
        bogie2Pivot.attachChild(bogie2);
        bogie2Pivot.setLocalTranslation(x, y, -z);

        //  Wheels
        x = vd.wheelPos.x;
        y = vd.wheelPos.y;
        z = vd.wheelPos.z;
        //  Bogie1
        Spatial wheels1 = loadModel("wheels");
        wheels1.setMaterial(matDebug);
        bogie1Pivot.attachChild(wheels1);
        wheels1.setLocalTranslation(x, y, z);
        wheels.add(wheels1);

        Spatial wheels2 = wheels1.clone();
        bogie1Pivot.attachChild(wheels2);
        wheels2.setLocalTranslation(x, y, -z);
        wheels.add(wheels2);
        //  Bogie2
        Spatial wheels3 = wheels1.clone();
        wheels3.setMaterial(matDebug);
        bogie2Pivot.attachChild(wheels3);
        wheels3.setLocalTranslation(x, y, z);
        wheels.add(wheels3);

        Spatial wheels4 = wheels1.clone();
        bogie2Pivot.attachChild(wheels4);
        wheels4.setLocalTranslation(x, y, -z);
        wheels.add(wheels4);

        //  Brakes
        x = vd.brakePos.x;
        y = vd.brakePos.y;
        z = vd.brakePos.z;
        //  Bogie1
        Spatial brakes1 = loadModel(vehicleType + "Brakes");
        //Spatial brakes1 = loadModel(assetManager, "wagonBrakes");
        brakes1.setMaterial(matDebug);
        bogie1Pivot.attachChild(brakes1);
        brakes1.setLocalTranslation(x, y, z);
        brakes.add(brakes1);

        Spatial brakes2 = brakes1.clone();
        bogie1Pivot.attachChild(brakes2);
        brakes2.setLocalTranslation(x, y, -z);
        brakes2.rotate(0, 180 * FastMath.DEG_TO_RAD, 0);
        brakes.add(brakes2);
        //  Bogie2
        Spatial brakes3 = brakes1.clone();
        bogie2Pivot.attachChild(brakes3);
        brakes3.setLocalTranslation(x, y, z);
        brakes.add(brakes3);

        Spatial brakes4 = brakes1.clone();
        bogie2Pivot.attachChild(brakes4);
        brakes4.setLocalTranslation(x, y, -z);
        brakes4.rotate(0, 180 * FastMath.DEG_TO_RAD, 0);
        brakes.add(brakes4);
    }

    protected Spatial loadModel(String name) {
        Spatial mesh;
        try {
            mesh = assetManager.loadModel("Models/Obj/" + name + ".j3o");
            return mesh;
        } catch (Exception e) {
            System.out.println("Couldn't find .j3o for " + name + ". Loading from .obj");
        }
        try {
            mesh = assetManager.loadModel("Models/Obj/" + name + ".obj");
            TangentBinormalGenerator.generate(mesh);
            return mesh;
        } catch (Exception e) {
            System.out.println("Couldn't find .obj for " + name + ". Loading from .blend");
        }
        try {
            mesh = assetManager.loadModel("Models/Obj/" + name + ".blend");
            return mesh;
        } catch (Exception e) {
            System.out.println("Couldn't find model for " + name + ". This is bad!");
        }
        return null;
    }

    protected Material loadMaterial(String name) {
        Material mat = new Material(assetManager, "Common/MatDefs/Light/Lighting.j3md");
        mat.setTexture("DiffuseMap", assetManager.loadTexture("Textures/Train/Link_m_" + name + "_BaseColor.png"));
        mat.setTexture("NormalMap", assetManager.loadTexture("Textures/Train/Link_m_" + name + "_Normal.png"));
        mat.setBoolean("UseMaterialColors", true);
        mat.setColor("Diffuse", ColorRGBA.White);  // minimum material color
        mat.setColor("Specular", ColorRGBA.White); // for shininess
        mat.setFloat("Shininess", 64f); // [1,128] for shininess

        return mat;
    }

    protected void addCoupling() {
        coupling1 = new Coupling(assetManager, this, 1);
        coupling2 = new Coupling(assetManager, this, 2);

        float x = vd.couplingPos.x;
        float y = vd.couplingPos.y;
        float z = vd.couplingPos.z;
        coupling1.getBasePivot().rotate(0, FastMath.DEG_TO_RAD * 180, 0);
        coupling1.getBasePivot().setLocalTranslation(x, y, -z);
        coupling2.getBasePivot().setLocalTranslation(x, y, z);

        if (/*Program.vrOn*/Program.loadDetailedComponents) {
            pivot.attachChild(coupling1.getBasePivot());
            pivot.attachChild(coupling2.getBasePivot());
        }
    }

    //adds air couplings
    protected void addAirCoupling() {
        air1 = new AirCoupling(assetManager, this, 1);
        air2 = new AirCoupling(assetManager, this, 2);

        float x = vd.airCouplingPos.x;
        float y = vd.airCouplingPos.y;
        float z = vd.airCouplingPos.z;
        air1.getBasePivot().rotate(0, FastMath.DEG_TO_RAD * 180, 0);
        air1.getBasePivot().setLocalTranslation(x, y, -z);
        air2.getBasePivot().setLocalTranslation(x, y, z);

        if (/*Program.vrOn*/Program.loadDetailedComponents) {
            pivot.attachChild(air1.getBasePivot());
            pivot.attachChild(air2.getBasePivot());
        }
        
        vbs = new VehicleBrakeSystem(air1, air2, true);
    }

    public Coupling getCoupling1() {
        return coupling1;
    }

    public Coupling getCoupling2() {
        return coupling2;
    }

    public Coupling getCoupling(int index) // Get coupling 0 or 1
    {
        if (index == 0) {
            return coupling1;
        } else {
            return coupling2;
        }
    }

    public AirCoupling getAirCoupling1() {
        return air1;
    }

    public AirCoupling getAirCoupling2() {
        return air2;
    }

    public AirCoupling getAirCoupling(int index) // Get air coupling 0 or 1
    {
        if (index == 0) {
            return air1;
        } else {
            return air2;
        }
    }
    
    public VehicleBrakeSystem getVehicleBrakeSystem() {
        return vbs;
    }

    public Guardrail getGuardrail1() {
        return guardrail1;
    }

    public Guardrail getGuardrail2() {
        return guardrail2;
    }

    public Guardrail getGuardrail(int index) // Get guardrail 0 or 1
    {
        if (index == 0) {
            return guardrail1;
        } else {
            return guardrail2;
        }
    }

    public void setSpolingOn(boolean spolingOn) {
        this.spolingOn = spolingOn;
    }

    public void setSpolingSpeedKmh(float spolingSpeedKmh) {
        this.spolingSpeedKmh = spolingSpeedKmh;
        spolingSpeedMs = spolingSpeedKmh / 3.6f;
    }
}

class VehicleData
{
    enum VehicleType {Locomotive, Wagon, LinkedWagon};
    
    public String name = "locomotive";
    public VehicleType vehicleType = VehicleType.Locomotive;
    
    public float colliderWidth = 1.4f;
    public float colliderLength = 6.6f;
    public float colliderLength2 = 7.3f;
    public float colliderHeight = 1.6f;
    public float trainColliderLength = 7.8f;
    
    public Vector3f guardrailPos = new Vector3f(0, 1.7f, 6.85f);
    public Vector3f couplingPos = new Vector3f(0, 1.26f, 7.31f);
    public Vector3f airCouplingPos = new Vector3f(-0.395f, 1.05f, 7.30f);
    public Vector3f bogiePos = new Vector3f(0, 0.0f, 3.15f);
    public float bogieOffset = .9f;
    public Vector3f wheelPos = new Vector3f(0, .68f, 1.22f);
    public Vector3f brakePos = new Vector3f(0, 0, 0);
    
    VehicleData(String name, VehicleType vehicleType, float colliderWidth, float colliderLength, float colliderLength2, float colliderHeight, float trainColliderLength, Vector3f guardrailPos, Vector3f couplingPos, Vector3f airCouplingPos, Vector3f bogiePos, float bogieOffset, Vector3f wheelPos, Vector3f brakePos)
    {
        this.name = name;
        this.vehicleType = vehicleType;
        
        this.colliderWidth = colliderWidth;
        this.colliderLength = colliderLength;
        this.colliderLength2 = colliderLength2;
        this.colliderHeight = colliderHeight;
        this.trainColliderLength = trainColliderLength;
    
        this.guardrailPos = guardrailPos;
        this.couplingPos = couplingPos;
        this.airCouplingPos = airCouplingPos;
        this.bogiePos = bogiePos;
        this.bogieOffset = bogieOffset;
        this.wheelPos = wheelPos;
        this.brakePos = brakePos;
    }
    
    VehicleData(String file)
    {
        readDataFromFile(file);
    }
    
    VehicleData()
    {
        
    }
    
    void readDataFromFile(String file)
    {
        
    }
    
    void writeDataToFile(String file)
    {
        
    }
}