/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package a_program;

import static a_program.Program.activeBalise;
import static a_program.Program.assetManager;
import static a_program.Program.bulletAppState;
import static a_program.Program.cam;
import static a_program.Program.simVeh;
import static a_program.Program.vrOn;
import static a_program.Program.vrVehicle;
import static a_program.Program.forwardMove;
import static a_program.Program.rightMove;
import static a_program.Program.upKey;
import static a_program.Program.leftKey;
import static a_program.Program.downKey;
import static a_program.Program.rightKey;
import static a_program.Program.editBaliseCodeIndex;
import static a_program.Program.crosshair;
import static a_program.Program.rootNode;
import static a_program.Program.settings;
import static a_program.Program.stVehicleList;
import static a_program.Program.vrAppState;
import com.jme3.app.VRAppState;
import com.jme3.asset.AssetManager;
import com.jme3.bullet.BulletAppState;
import com.jme3.bullet.collision.shapes.CompoundCollisionShape;
import com.jme3.bullet.control.BetterCharacterControl;
import com.jme3.bullet.control.RigidBodyControl;
import com.jme3.collision.CollisionResults;
import com.jme3.input.event.MouseButtonEvent;
import com.jme3.input.vr.VRInputType;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.FastMath;
import com.jme3.math.Matrix3f;
import com.jme3.math.Quaternion;
import com.jme3.math.Ray;
import com.jme3.math.Vector2f;
import com.jme3.math.Vector3f;
import com.jme3.renderer.Camera;
import com.jme3.scene.CameraNode;
import com.jme3.scene.Geometry;
import com.jme3.scene.Node;
import com.jme3.scene.Spatial;
import com.jme3.scene.control.CameraControl;
import com.jme3.scene.shape.Line;
import com.jme3.system.AppSettings;
import com.jme3.system.jopenvr.VRControllerAxis_t;
import com.jme3.system.jopenvr.VRControllerState_t;
import com.jme3.util.VRGUIPositioningMode;
import e_tco.GeometryAction;
import vehicle.AirCoupling;
import vehicle.Coupling;
import vehicle.Guardrail;
import vehicle.Vehicle;

/**
 *
 * @author Eirik
 */

//  Class for the person you move around in the VR simulator
public class VrPerson { 

    //---------------------------------------------------------
    //    VARIABLES
    //---------------------------------------------------------

    //  VR variables
    private BetterCharacterControl rb;                              //  Character physics controller
    private Node vrObserver;                                        //  Base Node for character
    private Node vrObserverBaseLok = new Node("vrobserverbaselok"); //  Node inside locomotive
    
    private int vrMainInputId = 1;                                  //  Input ID used for the main action of gripping objects, corresponding button/trigger may vary between VR systems
    private int vrSecondaryInputId = 2;                             //  Input ID used for secondary actions, corresponding button/trigger may vary between VR systems
    
    private Geometry leftHand, rightHand;                           //  Player hands
    private RigidBodyControl leftRB, rightRB;
    private Spatial rightHandLineGeo;                               //  Line for aiming
    boolean[] holding = new boolean[2];                             //  Whether a hand is holding something
    float[] prevGrip = new float[2];                                //  Grip value of the previous frame
    private boolean teleportMoveOnly = false;                         //  If true, allows the person to teleport rather than move with the "joystick"
    private boolean teleported = false;

    static boolean boarded = false;                                 //  Whether the character is on a train or not
    static Guardrail trainBoarded = null;                           //  The guardrail on the train that the character boarded

    
    //  NoVr variables
    private CameraNode noVrCamNodePerson;                           //  Camera Node for noVR
    private CameraNode noVrCamNodeLok;                              //  Camera Node for noVR inside locomotive
    boolean camRotated = true;                                      //  Used to check if camera was moved last frame

    private float noVrPersonRotateAlpha;                            //  Left/Right rotation
    private float noVrPersonRotateY;                                //  Up/Down rotation
    
    //  Are these really needed? Check usage in Program
    private boolean personModeVR = false;                           //  Person Mode
    private boolean noVrPersonMode = false;                         //  Person Mode

    
    //  Other variables
    private Vector3f personYRayLoc = new Vector3f();
    private Vector3f personYRayDir = new Vector3f(0, -1, 0);
    private CollisionResults personYResults = new CollisionResults();
    private Ray personRay = new Ray(personYRayLoc, personYRayDir);

    private CollisionResults personClickResults = new CollisionResults();
    private Ray personClickRay = new Ray(personYRayLoc, personYRayDir);

    Vector3f look = new Vector3f(1, 0, 0);                          //  Look direction vector

    private Vector3f[] vPrevArray = { new Vector3f(), new Vector3f() }; //  Previous Position of hands
    private Vector3f v = new Vector3f();

    
    
    //---------------------------------------------------------
    //    INIT
    //---------------------------------------------------------
    
    //  Vr Person constructor, all variables are required to interact with various parts of the program
    VrPerson() 
    {
        //  Create Base Node
        vrObserver = new Node("observer");

        //  Physics
        rb = new BetterCharacterControl(.2f, 1.2f, 1.0f);
        rb.setJumpForce(new Vector3f(0, 0, 0));
        rb.setGravity(new Vector3f(0, 9.81f, 0));
        vrObserver.addControl(rb);
        bulletAppState.getPhysicsSpace().add(rb);
        
        //  Initialize VR and noVR variables
        if (vrOn) 
        {
            //  Initialize VR aspects
            vrAppState.setObserver(vrObserver);
            vrObserverBaseLok.attachChild(vrObserver);
            
            initHands();

            vrAppState.getVRGUIManager().setPositioningMode(VRGUIPositioningMode.AUTO_OBSERVER_ALL);
            //vrAppState.getVRGUIManager().setPositioningMode(VRGUIPositioningMode.AUTO_CAM_ALL_SKIP_PITCH);
            vrAppState.getVRGUIManager().setGuiScale(1.0f);
            vrAppState.getVRGUIManager().setPositioningElasticity(10f);

            vrLoop();
            toggleCamVR();
        } 
        else 
        {
            //  Create Camera nodes
            noVrCamNodePerson = new CameraNode("Camera person", cam);
            noVrCamNodePerson.setControlDir(CameraControl.ControlDirection.SpatialToCamera);
            noVrCamNodePerson.setLocalTranslation(0, 1.25f, 0);
            vrObserver.attachChild(noVrCamNodePerson);
            rootNode.attachChild(vrObserver);

            noVrCamNodeLok = new CameraNode("Camera Node", cam);
            noVrCamNodeLok.setControlDir(CameraControl.ControlDirection.SpatialToCamera);
            noVrCamNodeLok.setEnabled(true);

            //  Aim line geo
            Line line = new Line(new Vector3f(0, 0, 0), new Vector3f(0, 0, 50));
            Spatial lineGeo = new Geometry("Line", line);
            Material mat = new Material(
                    assetManager, "Common/MatDefs/Misc/Unshaded.j3md");
            mat.setColor("Color", ColorRGBA.Red);
            lineGeo.setMaterial(mat);
            lineGeo.setLocalTranslation(0, 1.25f, 0);
            //vrObserver.attachChild(lineGeo);
            

            //  Sets up crosshair
            crosshair.setImage(assetManager, "Textures/yellow.png", true);
            crosshair.setWidth(settings.getHeight() / 200);
            crosshair.setHeight(settings.getHeight() / 200);
            crosshair.setPosition(settings.getWidth() / 2, settings.getHeight() / 2);
        }
    }


    //---------------------------------------------------------
    //    UPDATE LOOP
    //---------------------------------------------------------
    
    public void update() 
    {
        if (vrOn) 
        {
            vrLoop();
        } 
        else 
        {
            noVrLoop();
        }
    }

    public void vrLoop() 
    {
        //  Update VR appstate
        vrAppState.getVRGUIManager().setPositioningMode(VRGUIPositioningMode.AUTO_OBSERVER_ALL);
        //vrAppState.getVRGUIManager().setPositioningMode(VRGUIPositioningMode.AUTO_CAM_ALL_SKIP_PITCH);
        vrAppState.getVRGUIManager().setGuiScale(1.0f);
        vrAppState.getVRGUIManager().setPositioningElasticity(10f);

        //  Move VR person
        if (personModeVR) 
        {
            moveAroundVR(1);

            handleWandInput(0, rightHand, rightRB);
            handleWandInput(1, leftHand, leftRB);
        }

        //  Move aim geo
        rightHandLineGeo.setLocalRotation(rightRB.getPhysicsRotation());
        rightHandLineGeo.setLocalTranslation(rightRB.getPhysicsLocation());
    }

    public void noVrLoop() 
    {
        if (noVrPersonMode) 
        {
            if (camRotated)                 //  If camera was moved last frame
            {
                camRotated = false;
            } 
            else 
            {
                noVrPersonRotateAlpha = 0;  //  stop rotation
                noVrPersonRotateY = 0;      //  stop rotation
            }
            noVrPersonMoveAround();         //  Move person
        }
    }

    
    //---------------------------------------------------------
    //    VR FUNCTIONS
    //---------------------------------------------------------
    private void initHands() 
    {
        //  Create hand wands
        leftHand = (Geometry) assetManager.loadModel("Models/vive_controller.j3o");
        rightHand = leftHand.clone();
        Material handMat = new Material(assetManager, "Common/MatDefs/Misc/Unshaded.j3md");
        handMat.setTexture("ColorMap", assetManager.loadTexture("Textures/vive_controller.png"));
        leftHand.setMaterial(handMat);
        rightHand.setMaterial(handMat);
        vrObserver.attachChild(rightHand);
        vrObserver.attachChild(leftHand);
        
        CompoundCollisionShape shape = new CompoundCollisionShape(); // Empty shape allows hands to move through anything
        
        leftRB = new RigidBodyControl(shape, 0);
        leftHand.addControl(leftRB);
        bulletAppState.getPhysicsSpace().add(leftRB);
        
        rightRB = new RigidBodyControl(shape, 0);
        rightHand.addControl(rightRB);
        bulletAppState.getPhysicsSpace().add(rightRB);

        //  Aim line geo
        Line line = new Line(new Vector3f(0, 0, 0), new Vector3f(0, 0, 5));
        rightHandLineGeo = new Geometry("Line", line);
        Material mat = new Material(assetManager, "Common/MatDefs/Misc/Unshaded.j3md");
        mat.setColor("Color", ColorRGBA.Red);
        rightHandLineGeo.setMaterial(mat);
        rootNode.attachChild(rightHandLineGeo);
    }

    //  Toggles the VR Camera and places the person next to the sim vehicle
    public void toggleCamVR() 
    {
        System.out.println("togglecam");
        personModeVR = true;
        //    simVeh.getPivot().detachChild(observer);
        vrObserverBaseLok.detachChild(vrObserver);
        rootNode.attachChild(vrObserver);
        if (simVeh != null)
        {
            Vector3f v = simVeh.getCoupling2().getBasePivot().getWorldTranslation();
            v = v.add(simVeh.getCoupling2().getBasePivot().getWorldRotation().mult(new Vector3f(2, 0, 2)));
            Vector3f tr1vec = simVeh.getTrackRunner1().getPoint();
            v.y = tr1vec.y + 0.2f;
            rb.warp(v);
        }
    }

    private void moveHands1Frame(int index) 
    {
        v = vPrevArray[index];
        vPrevArray[index] = vrAppState.getVRinput().getFinalObserverPosition(index);
    }

    private void handleWandInput(int index, Geometry geo, RigidBodyControl rb) 
    {
        if (vrAppState.getVRinput() == null || vrAppState.getVRinput().getFinalObserverPosition(index) == null) 
            return;
        

        Quaternion q = vrAppState.getVRinput().getFinalObserverRotation(index);
        //Vector3f v = vrAppState.getVRinput().getFinalObserverPosition(index);

        moveHands1Frame(index); // nødvendig dessverre pga 1 frame forflyttinga som er blitt oppdaga

        //  Move hands
        geo.setCullHint(Spatial.CullHint.Dynamic); // make sure we see it
        rb.setPhysicsLocation(v);
        rb.setPhysicsRotation(q);
        /*
        geo.setLocalTranslation(v.subtract(vrObserver.getWorldTranslation()));
        geo.setLocalRotation(q);
        */
        
        if (index == 0)
        {
            if (getButton(index)) 
            {
                if (!teleported)
                {
                    pointingAtVR(rb.getPhysicsLocation(), rb.getPhysicsRotation());
                    teleported = true;
                }
            }
            else
                teleported = false;
        }
        

        //  Grip, trigger
        float grip = getAxis(index, vrMainInputId);
        float holdRange = .2f;
        
        if (grip > 0.8f) 
        {
            vrVehicle.vrGripBrems(v);
            vrVehicle.vrGripPadr(v);
        }

        //  Interact with objects through grip
        if (grip < 0.5f && prevGrip[index] > 0.5f) 
        {
            for (Vehicle veh : stVehicleList) 
            {
                for (int i = 0; i < 2; i++) 
                {
                    //  Handles
                    if (veh.getCoupling(i).getHandlePos().distance(rb.getPhysicsLocation()) < holdRange) 
                        veh.getCoupling(i).interactMidOffset();

                    //  Boarding with Guardrails
                    if (veh.getGuardrail(i) != null)
                        if (veh.getGuardrail(i).getGuardrailPos().distance(rb.getPhysicsLocation()) < holdRange) 
                            veh.getGuardrail(i).click();

                    //  Air Valves
                    if (veh.getAirCoupling(i).getAirValvePivot().getWorldTranslation().distance(rb.getPhysicsLocation()) < holdRange)
                        veh.getAirCoupling(i).airValveInteract();
                }
            }
        }
        prevGrip[index] = grip;

        //----------------------------------------------
        //  Couplings and Air Couplings
        //----------------------------------------------
        
        //  Grabbing couplings
        if (grip > 0.5f && !holding[index]) 
        {
            //pointingAtVR(v, q);

            //  Coupling
            outerloop:
            for (Vehicle veh : stVehicleList) 
            {
                for (int i = 0; i < 2; i++) 
                {
                    if (veh.getCoupling(i).heldBy() == null) 
                    {
                        // Couplings
                        if (veh.getCoupling(i).getEndPivot().getWorldTranslation().distance(rb.getPhysicsLocation()) < holdRange) 
                        {
                            veh.getCoupling(i).hold(geo);
                            holding[index] = true;
                            break outerloop;
                        } 
                        else if (veh.getCoupling(i).getState() != Coupling.CouplingState.Connected) 
                        {
                            veh.getCoupling(i).setState(Coupling.CouplingState.Loose);
                        }
                    }

                    if (veh.getAirCoupling(i).heldBy() == null) 
                    {
                        // Air Couplings
                        if (veh.getAirCoupling(i).getEndPivot().getWorldTranslation().distance(rb.getPhysicsLocation()) < holdRange) 
                        {
                            veh.getAirCoupling(i).hold(geo);
                            holding[index] = true;
                            break outerloop;
                        } 
                        else if (veh.getAirCoupling(i).getState() != Coupling.CouplingState.Connected) 
                        {
                            veh.getAirCoupling(i).setState(Coupling.CouplingState.Loose);
                        }
                    }
                }
            }
        }
        
        //  Releasing and connecting couplings
        if (grip <= 0.5f && holding[index]) 
        {
            for (Vehicle veh : stVehicleList) 
            {
                for (Vehicle veh2 : stVehicleList) 
                {
                    if (veh == veh2) // Don't check vehicle with itself
                        continue;
                    

                    for (int i = 0; i < 2; i++) 
                    {
                        if (veh.getCoupling(i).heldBy() == geo) 
                        {
                            //  Check Coupling on other vehicle
                            for (int j = 0; j < 2; j++) 
                            {
                                if (veh.getCoupling(i).getEndPivot().getWorldTranslation().distance(
                                        veh2.getCoupling(j).getConnectPivot().getWorldTranslation()
                                ) < holdRange) 
                                {
                                    //  Connect
                                    veh.getCoupling(i).connectTo(veh2.getCoupling(j));
                                    holding[index] = false;
                                }
                            }
                            if (veh.getCoupling(i).getState() != Coupling.CouplingState.Connected) 
                            {
                                veh.getCoupling(i).release();
                                holding[index] = false;
                            }
                        }

                        if (veh.getAirCoupling(i).heldBy() == geo) 
                        {
                            //  Check Air Coupling on other vehicle
                            for (int j = 0; j < 2; j++) 
                            {
                                if (veh.getAirCoupling(i).getEndPivot().getWorldTranslation().distance(
                                        veh2.getAirCoupling(j).getConnectPivot().getWorldTranslation()
                                ) < holdRange) 
                                {
                                    //  Connect
                                    veh.getAirCoupling(i).connectTo(veh2.getAirCoupling(j));
                                    holding[index] = false;
                                }
                            }
                            if (veh.getAirCoupling(i).getState() != Coupling.CouplingState.Connected) 
                            {
                                veh.getAirCoupling(i).release();
                                holding[index] = false;
                            }
                        }
                    }
                }
            }
        }
    }
    
    //connects air
    public static void airConnect(AirCoupling ac) {
        for (Vehicle veh : stVehicleList) {
            // connection 1
            if (veh.getAirCoupling1() != ac) {
                float distance = veh.getAirCoupling1().getBasePivot().getWorldTranslation().distance(
                        ac.getBasePivot().getWorldTranslation());

                if (distance < 1) {
                    ac.connectTo(veh.getAirCoupling1());
                }
            }

            // connection 2
            if (veh.getAirCoupling2() != ac) {
                float distance = veh.getAirCoupling2().getBasePivot().getWorldTranslation().distance(
                        ac.getBasePivot().getWorldTranslation());

                if (distance < 1) {
                    ac.connectTo(veh.getAirCoupling2());
                }
            }
        }
    }

    //connects trains
    public static void connect(Coupling c) {
        for (Vehicle veh : stVehicleList) {
            // coupling 1
            if (veh.getCoupling(0) != c) {
                float distance = veh.getCoupling(0).getBasePivot().getWorldTranslation().distance(
                        c.getBasePivot().getWorldTranslation());

                if (distance < 1) {
                    c.connectTo(veh.getCoupling(0));
                }
            }

            // coupling 2
            if (veh.getCoupling(1) != c) {
                float distance = veh.getCoupling(1).getBasePivot().getWorldTranslation().distance(
                        c.getBasePivot().getWorldTranslation());

                if (distance < 1) {
                    c.connectTo(veh.getCoupling(1));
                }
            }
        }
    }

    private void pointingAtVR(Vector3f startVec, Quaternion dirQuat) 
    {
        CollisionResults results = new CollisionResults();
        Ray ray = new Ray(startVec, dirQuat.getRotationColumn(2).mult(1).normalize());
        rootNode.collideWith(ray, results);
        //System.out.println("Pointing at closest:");

        if (results.getClosestCollision() != null && results.getClosestCollision().getGeometry() instanceof GeometryAction) 
        {
            //System.out.println("geometry action");
            ((GeometryAction) results.getClosestCollision().getGeometry()).geometryClicked();
        }
        
        // teleport forflytt
        if (    results.getClosestCollision() != null
                && results.getClosestCollision().getGeometry().getName().contains("batch")
                && !(results.getClosestCollision().getGeometry() instanceof GeometryAction)) 
        {
            rb.warp(results.getClosestCollision().getContactPoint());
        }

        printCollision(results);
    }

    public void noVrPersonModeMouseClick(MouseButtonEvent evt) 
    {
        personClickResults.clear();
        Vector2f click2d = new Vector2f(settings.getWidth() / 2, settings.getHeight() / 2);
        Vector3f click3d = cam.getWorldCoordinates(new Vector2f(click2d.x, click2d.y), 0).clone();
        Vector3f dir = cam.getWorldCoordinates(new Vector2f(click2d.x, click2d.y), 1f).subtractLocal(click3d).normalizeLocal();

        personClickRay.setOrigin(click3d);
        personClickRay.setDirection(dir);
        rootNode.collideWith(personClickRay, personClickResults);

        // fjern evt merking av balise
        if (activeBalise != null) 
        {
            activeBalise.unMark();
            editBaliseCodeIndex = 0;
        }

        // finn ut om et klikkbart objekt er truffet
        if (personClickResults.getClosestCollision() != null && personClickResults.getClosestCollision().getGeometry() instanceof GeometryAction) 
        {
            System.out.println("geometry action");
            ((GeometryAction) personClickResults.getClosestCollision().getGeometry()).geometryClicked();
        }

        for (int i = 0; i < personClickResults.size(); i++) 
        {
            float dist = personClickResults.getCollision(i).getDistance();
            Vector3f pt = personClickResults.getCollision(i).getContactPoint();
            String target = personClickResults.getCollision(i).getGeometry().getName();
            Geometry geom = personClickResults.getCollision(i).getGeometry();
            System.out.println("Selection #" + i + ": " + target + " at " + pt + ", " + dist + " WU away." + " orig: " + click3d + " dir: " + dir);
        }
    }
    
    public void noVrPersonModeTeleport(MouseButtonEvent evt) 
    {
        personClickResults.clear();
        Vector2f click2d = new Vector2f(settings.getWidth() / 2, settings.getHeight() / 2);
        Vector3f click3d = cam.getWorldCoordinates(new Vector2f(click2d.x, click2d.y), 0).clone();
        Vector3f dir = cam.getWorldCoordinates(new Vector2f(click2d.x, click2d.y), 1f).subtractLocal(click3d).normalizeLocal();

        personClickRay.setOrigin(click3d);
        personClickRay.setDirection(dir);
        rootNode.collideWith(personClickRay, personClickResults);

        // fjern evt merking av balise
        if (activeBalise != null) 
        {
            activeBalise.unMark();
            editBaliseCodeIndex = 0;
        }

        // teleport forflytt
        if (    personClickResults.getClosestCollision() != null
                && personClickResults.getClosestCollision().getGeometry().getName().contains("batch")
                && !(personClickResults.getClosestCollision().getGeometry() instanceof GeometryAction)) 
        {
            rb.warp(personClickResults.getClosestCollision().getContactPoint());
        }
        
        for (int i = 0; i < personClickResults.size(); i++) 
        {
            float dist = personClickResults.getCollision(i).getDistance();
            Vector3f pt = personClickResults.getCollision(i).getContactPoint();
            String target = personClickResults.getCollision(i).getGeometry().getName();
            Geometry geom = personClickResults.getCollision(i).getGeometry();
            System.out.println("Selection #" + i + ": " + target + " at " + pt + ", " + dist + " WU away." + " orig: " + click3d + " dir: " + dir);
        }
    }
    
    private void printCollision(CollisionResults results) 
    {
        for (int i = 0; i < results.size(); i++) 
        {
            float dist = results.getCollision(i).getDistance();
            Vector3f pt = results.getCollision(i).getContactPoint();
            String target = results.getCollision(i).getGeometry().getName();
            Geometry geom = results.getCollision(i).getGeometry();

            if (dist < 5) 
            {
                System.out.println("COLLISIONdist : " + dist + " pt" + pt + " target: " + target + " geom: " + geom);
            }
        }
    }

    private float getAxis(int index, int axis) 
    {
        VRControllerState_t rawstate = (VRControllerState_t) vrAppState.getVRinput().getRawControllerState(index);
        VRControllerAxis_t cs = rawstate.rAxis[axis];
        return cs.x;
    }
    
    private boolean getButton(int index) 
    {
        return vrAppState.getVRinput().isButtonDown(index, VRInputType.ViveGripButton); //Button needs to be changed to work with Oculus!
    }
    
    public void board(Guardrail rail) 
    {
        boarded = !boarded;
        if (boarded) 
        {
            trainBoarded = rail;
        } 
        else 
        {
            Vector3f pos = trainBoarded.getGuardrailPos();
            Vector3f v = new Vector3f(2f, -1f, 0);
            pos = pos.add( trainBoarded.getVehicle().getPivot().getWorldRotation().mult(v) );
            warp(pos);
            trainBoarded = null;
        }
        System.out.println("Boarded: " + boarded);
    }

    public void warp(Vector3f p) 
    {
        rb.warp(p);
    }

    private void moveAroundVR(int index) 
    {
        float r = -vrAppState.getVRinput().getAxis(index, VRInputType.ViveTrackpadAxis).x;
        float f = vrAppState.getVRinput().getAxis(index, VRInputType.ViveTrackpadAxis).y;

        moveAround(f, r, vrObserver.getLocalRotation().inverse().mult(vrAppState.getFinalObserverRotation()), vrObserver);
    }

    private void moveAround(float f, float r, Quaternion rotation, Spatial node) 
    {
        float speed = 4.0f;

        // Move
        if (boarded) {
            rb.warp(trainBoarded.getGuardrailPos());
        } else {
            if (teleportMoveOnly)
                return;
            
            Vector3f move = node.getLocalRotation().mult(rotation.mult(new Vector3f(r, 0, f)));
            move.setY(0);
            move = move.normalize().mult(new Vector2f(f, r).length()).mult(speed);

            rb.setWalkDirection(move);
        }
    }

    private void noVrPersonMoveAround() 
    {
        // flytt
        forwardMove = upKey - downKey;
        rightMove = leftKey - rightKey;

        float f = forwardMove;
        float r = rightMove;
        float d = noVrPersonRotateAlpha;
        float y = noVrPersonRotateY;
        //System.out.println(d + ", " + y);

        moveAround(f, r, vrObserver.getLocalRotation().inverse().mult(vrObserver.getLocalRotation()), vrObserver);

        //  Rotation
        //if (d > 0.1f || d < -0.1f || y > 0.1f || y < -0.1f)
        {
            float tpf = Vehicle.secsSinceLastFrame;
            float grad = d * 8f / 60f;
            float gradY = y * .5f / 60f;

            //rotation matrix for the x axis
            Matrix3f rotX = new Matrix3f();
            rotX.fromAngleAxis(FastMath.DEG_TO_RAD * grad * 5f, new Vector3f(0f, 1f, 0f));
            look = rotX.mult(look);

            //vector pointing right
            Vector3f right = look.cross(new Vector3f(0f, 1f, 0f));

            //rotation matrix for the y axis
            Matrix3f rotY = new Matrix3f();
            rotY.fromAngleAxis(FastMath.DEG_TO_RAD * gradY * 5f, right);
            look = rotY.mult(look);

            //cap Y of look direction
            if (look.y > 0.8f) {
                look.setY(0.8f);
            } else if (look.y < -0.8f) {
                look.setY(-0.8f);
            }

            look = look.normalize();

            //look at a point in front of the person according to look direction
            //person.lookAt(person.getWorldTranslation().add(look), new Vector3f(0f,1f,0f));
            rb.setViewDirection(look);
            noVrCamNodePerson.rotate(gradY, 0, 0);
        }
    }
    
    //----------------------------------------------------
    //  Get/Set functions
    //----------------------------------------------------

    public void setNoVrPersonRotateAlpha(float a) {
        noVrPersonRotateAlpha = a;
    }

    public void setNoVrPersonRotateY(float y) {
        noVrPersonRotateY = y;
    }

    public void setVrObserverBaseLok(Quaternion r, Vector3f p) {
        vrObserverBaseLok.setLocalRotation(r);
        vrObserverBaseLok.setLocalTranslation(p);
    }

    public boolean getNoVrPersonMode() {
        return noVrPersonMode;
    }

    public void setNoVrPersonMode(boolean bool) {
        noVrPersonMode = true;
    }

    public CameraNode getNoVrCamNodeLok() {
        return noVrCamNodeLok;
    }

    public CameraNode getNoVrCamNodePerson() {
        return noVrCamNodePerson;
    }

    public void setNoVrCamNodeLok(boolean bool) {
        noVrCamNodeLok.setEnabled(bool);
    }

    public void setNoVrCamNodePerson(boolean bool) {
        noVrCamNodePerson.setEnabled(bool);
    }

    public void setCamRotated(boolean bool) {
        camRotated = bool;
    }
    
    public Vector3f getPos()
    {
        return vrObserver.getWorldTranslation();
    }
    
    public void setPos(Vector3f pos)
    {
        warp(pos);
    }
    
    public Vector3f getRot()
    {
        return look;
    }
    
    public void setRot(Vector3f rot)
    {
        look = rot;
    }
    
    public void setTeleportActive(boolean b)
    {
        teleportMoveOnly = b;
    }
    
    public boolean getVrOn()
    {
        return vrOn;
    }
}
